/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package io.fusion.fusionbackend.model;

import lombok.Getter;
import lombok.NoArgsConstructor;
import lombok.Setter;
import lombok.experimental.SuperBuilder;

import javax.persistence.Column;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.MappedSuperclass;
import javax.persistence.Version;
import java.io.Serializable;
import java.util.Objects;

@MappedSuperclass
@Getter
@Setter
@SuperBuilder
@NoArgsConstructor
public abstract class BaseEntity implements Serializable {
    @Id
    @Column(updatable = false, nullable = false)
    @GeneratedValue(strategy = GenerationType.SEQUENCE, generator = "idgen")
    private Long id;

    @Version
    private Long version;

    protected void copyFrom(BaseEntity sourceBaseEntity) {

        if (sourceBaseEntity.getId() != null) {
            setId(sourceBaseEntity.getId());
        }

        if (sourceBaseEntity.getVersion() != null) {
            setVersion(sourceBaseEntity.getVersion());
        }
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (!(o instanceof BaseEntity)) {
            return false;
        }
        BaseEntity that = (BaseEntity) o;
        if (id == null) {
            return false;
        }
        return Objects.equals(id, that.id);
    }

    /* Fixing the entity identifier equals and hashCode
        The problem with id as hashcode:
        When the entity was first stored in the Set, the identifier was null. After the entity was persisted,
        the identifier was assigned to a value that was automatically generated, hence the hashCode differs.
        For this reason, the entity cannot be found in the Set after it got persisted.
        https://vladmihalcea.com/how-to-implement-equals-and-hashcode-using-the-jpa-entity-identifier/
     */

    @Override
    public int hashCode() {
        return getClass().hashCode();
    }

    @Override
    public String toString() {
        return "BaseEntity{"
                + "id=" + id
                + ", version=" + version
                + '}';
    }
}
